#include <gtest/gtest.h>
#include "util.h"

#define TEST_COMPILE(src) DO(test_compile(src, "verify"))
#define TEST_ERROR(src) DO(test_error(src, "verify"))

class StableTypeTest : public PassTest
{};

TEST_F(StableTypeTest, EphemeralIsoReturnShouldCompile)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    None\n"
    "  fun good(): A iso^ =>\n"
    "    recover A end\n";

  TEST_COMPILE(src);
}

TEST_F(StableTypeTest, IncompatibleIsectReturnShouldCompile)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    None\n"
    "  fun good(): (A ref & A val) =>\n"
    "    recover A end\n";

  TEST_COMPILE(src);
}

TEST_F(StableTypeTest, GenericSelfIsectShouldCompile)
{
  const char* src =
    "class Generic[A]\n"
    "  fun doit(a: A) =>\n"
    "    let x: (A & A) = consume a\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_COMPILE(src);
}

TEST_F(StableTypeTest, CompatibleIsectLetField)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  let x: (A box & A val) = recover A end\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_COMPILE(src);
}

TEST_F(StableTypeTest, EphemeralIsoEmbedField)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  embed x: A iso^ = recover A end\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, IncompatibleIsectLetField)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  let x: (A ref & A val) = recover A end\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralIsoLetField)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  let x: A iso^ = recover A end\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralIsoVarField)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  var x: A iso^ = recover A end\n"
    "  new create(env: Env) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, IncompatibleIsectLet)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let x: (A ref & A val) = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, IncompatibleIsectVar)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    var x: (A ref & A val) = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralTrnLet)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let x: A trn^ = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralTrnVar)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    var x: A trn^ = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralIsoLet)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    let x: A iso^ = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralIsoVar)
{
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    var x: A iso^ = recover A end\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, EphemeralIsoParameter)
{
  // parameters should reject even if never called
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    None\n"
    "  fun bad(a: A iso^) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, IncompatibleIsectParameter)
{
  // parameters should reject even if never called
  const char* src =
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    None\n"
    "  fun bad(a: (A ref & A val)) =>\n"
    "    None\n";

  TEST_ERROR(src);
}

TEST_F(StableTypeTest, DoubleEphemeralInstantiation)
{
  const char* src =
    "class Generic[T]\n"
    "  fun get(f: {(): T^}): T =>\n"
    "    f()\n"
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    Generic[A iso^].get({() => A})\n";

  TEST_COMPILE(src);
}

TEST_F(StableTypeTest, EphemeralInstantiationRespectsEphemeralUnions)
{
  const char* src =
    "type Arg is (A iso | None)\n"
    "class Generic[T]\n"
    "  fun get(f: {(): T^}): T =>\n"
    "    f()\n"
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    Generic[Arg^].get({() => A})\n";

  TEST_COMPILE(src);
}
TEST_F(StableTypeTest, EphemeralInstantiationRespectsUnionsOfEphemerals)
{
  const char* src =
    "type Arg is (A iso^ | None)\n"
    "class Generic[T]\n"
    "  fun get(f: {(): T^}): T =>\n"
    "    f()\n"
    "class A\n"
    "actor Main\n"
    "  new create(env: Env) =>\n"
    "    Generic[Arg].get({() => A})\n";

  TEST_COMPILE(src);
}
